@@ -7,30 +7,12 @@ from django.db.models import Count, Sum, Max, Min
 from django.db.models.functions import TruncMonth, TruncDay
 from django.core.urlresolvers import reverse
 from django.conf import settings
-from django.dispatch import receiver
-from django.db.models.signals import post_save
 from django.apps import apps
 from django.utils import timezone
 
 from .dt import get_current_week_range_datetime, get_now, LOCAL_TIMEZONE
 
 
-# 更新任务
-def update_user_missions(user, stage_list):
-    for stage in stage_list:
-        mission_list = Mission.objects.filter(stage=stage)
-        for mission in mission_list:
-            # 检查重复，没有则创建
-            try:
-                UserMission.objects.get(user=user, mission=mission)
-                continue
-            except:
-                user_mission = UserMission()
-                user_mission.user = user
-                user_mission.mission = mission
-                user_mission.save()
-
-
 # Create your models here.
 
 class ContentType(models.Model):
@@ -43,7 +25,7 @@ class ContentType(models.Model):
         db_table = 'bee_django_mission_content_type'
         app_label = 'bee_django_mission'
         ordering = ['id']
-        unique_together=("app_label",'model')
+        unique_together = ("app_label", 'model')
 
     def __str__(self):
         return self.app_label + '.' + self.model
@@ -64,6 +46,9 @@ class Line(models.Model):
         app_label = 'bee_django_mission'
         ordering = ['id']
         verbose_name = '任务线'
+        permissions = (
+            ('can_manage_mission', '可以进入mission管理页'),
+        )
 
     def __str__(self):
         return self.name
@@ -71,10 +56,10 @@ class Line(models.Model):
     def __unicode__(self):
         return self.name
 
-    # 获取下一个stage，如stage为空，则获取第一个stage
-    def get_next_stage(self, last_stage=None):
-        if last_stage:
-            current_stage_level = last_stage.level
+    # 获取下一个stage，如stage为空，则获取第一个stage，没有返回空
+    def get_next_stage(self, current_stage=None):
+        if current_stage:
+            current_stage_level = current_stage.level
         else:
             current_stage_level = 0
         next_stage_list = self.stage_set.filter(level__gt=current_stage_level).order_by("level")
@@ -84,42 +69,11 @@ class Line(models.Model):
         else:
             return None
 
+    def is_week_line(self):
+        return self.line_type == 2
 
-            # def add_user_first_stage(self, user):
-            #     # 已经有未完成的stage
-            #     user_stage_list = UserStage.objects.filter(user=user, finish_at__isnull=True)
-            #     if user_stage_list.count() > 0:
-            #         return
-            #
-            #     # 找到第一个任务块
-            #     stage_list = self.stage_set.all().order_by("level")
-            #     if stage_list.count() > 0:
-            #         stage = stage_list.first()
-            #     else:
-            #         return
-            #
-            #     # 添加
-            #     u_s = UserStage()
-            #     u_s.user = user
-            #     u_s.stage = stage
-            #     u_s.save()
-            #     return
-            #
-            # def add_user_next_stage(self, user):
-            #     # 已经有未完成的stage
-            #     user_stage_list = UserStage.objects.filter(user=user, finish_at__isnull=True)
-            #     if user_stage_list.count() > 0:
-            #         return
-            #
-            #     # 找到当前最高stage
-            #     user_stage_list = UserStage.objects.filter(user=user, finish_at__isnull=False).order_by('-stage__level')
-            #     if user_stage_list.count() > 0:
-            #         user_stage = user_stage_list.first()
-            #         max_stage_level = user_stage.level
-            #     else:
-            #         return
-            #
-            #     stage_list = self.stage_set.filter().order_by("level")
+    def is_unlimited_line(self):
+        return self.line_type == 1
 
 
 class Stage(models.Model):
@@ -135,39 +89,30 @@ class Stage(models.Model):
         ordering = ['level']
         verbose_name = '阶段任务'
 
+    def get_absolute_url(self):
+        return reverse('bee_django_mission:stage_unlimited_list')
+
     def __str__(self):
         return self.line.name + "-" + self.name
 
     def __unicode__(self):
         return self.line.name + "-" + self.name
 
-        # def add_to_user_stage(self, user):
-        #     user_stage = UserStage()
-        #     user_stage.user = user
-        #     user_stage.stage = self
-        #     user_stage.save()
-        #
-        # def update_user_mission(self, user):
-        #     mission_list = self.mission_set.all()
-        #     for mission in mission_list:
-        #         # 检查重复，没有则创建
-        #         try:
-        #             UserMission.objects.get(user=user, mission=mission)
-        #             continue
-        #         except:
-        #             user_mission = UserMission()
-        #             user_mission.user = user
-        #             user_mission.mission = mission
-        #             user_mission.save()
+    @classmethod
+    def get_week_stage(cls):
+        try:
+            return cls.objects.get(line__line_type=2)
+        except:
+            return
 
 
 MISSION_AGGREGATE_TYPE_CHOICES = ((1, "Count"), (2, "Sum"), (3, "TruncDay"))
 MISSION_COMPARISON_TYPE_CHOICES = ((1, '>='), (2, '>'))
-MISSION_OPERATOR_TYPE_CHOICES = ((0,'无'),(1, '* 60'),)
+MISSION_OPERATOR_TYPE_CHOICES = ((0, '无'), (1, '* 60'),)
 
 
 class MissionType(models.Model):
-    name = models.CharField(max_length=180, verbose_name='标题',unique=True)
+    name = models.CharField(max_length=180, verbose_name='标题', unique=True)
     content_type = models.ForeignKey(ContentType, on_delete=models.SET_NULL, null=True, verbose_name='app及model')
     aggregate_type = models.IntegerField(default=1, choices=MISSION_AGGREGATE_TYPE_CHOICES,
                                          verbose_name='聚合类型')  # max/count/sum
@@ -197,10 +142,10 @@ class MissionType(models.Model):
 
 
 class Mission(models.Model):
-    mission_type = models.ForeignKey(MissionType, on_delete=models.SET_NULL, null=True)
-    stage = models.ForeignKey(Stage, on_delete=models.SET_NULL, null=True)
+    mission_type = models.ForeignKey(MissionType, on_delete=models.SET_NULL, null=True, verbose_name='任务类型')
+    stage = models.ForeignKey(Stage, on_delete=models.SET_NULL, null=True, verbose_name='所属阶段')
     name = models.CharField(max_length=180, verbose_name='标题')
-    count = models.IntegerField(verbose_name='数量')
+    count = models.IntegerField(verbose_name='要求数量')
     info = models.TextField(verbose_name='备注', null=True, blank=True)
     created_at = models.DateTimeField(auto_now_add=True)
     order_by = models.IntegerField(verbose_name='顺序', null=True, blank=True)
@@ -211,6 +156,9 @@ class Mission(models.Model):
         ordering = ['created_at']
         verbose_name = '任务'
 
+    def get_absolute_url(self):
+        return reverse('bee_django_mission:mission_list')
+
     def __str__(self):
         return self.name
 
@@ -236,48 +184,76 @@ class UserLine(models.Model):
     def __unicode__(self):
         return self.line.name
 
-    # 获取该学生，同一user_line下，最后完成的user_stage
-    def get_last_finished_user_stage(self):
-        finished_user_stage_list = UserStage.objects.filter(user_line=self, finish_at__isnull=False).order_by(
-            'finish_at')
-        if finished_user_stage_list.count() > 0:
-            return finished_user_stage_list.last()
+    # 获取该学生，同一user_line下，最后的user_stage，没有返回空
+    def get_last_user_stage(self, status_list=None):
+        user_stage_list = UserStage.objects.filter(user_line=self).order_by(
+            'created_at')
+        if status_list:
+            user_stage_list = user_stage_list.filter(stutas__in=status_list)
+        if user_stage_list.count() > 0:
+            return user_stage_list.last()
         else:
             return None
 
     # 获取该学生，同一user_line下的所有stage
-    def get_all_user_stage(self):
+    def get_user_all_stage(self):
         user_stage_list = UserStage.objects.filter(user_line=self).order_by(
-            'finish_at')
+            'created_at')
         return user_stage_list
 
-    # 获取该学生，同一user_line下，正在进行的user_stage
+    # 获取该学生，同一user_line下，正在进行的user_stage，没有返回空
     def get_woking_user_stage(self):
-        woking_user_stage_list = UserStage.objects.filter(user_line=self, finish_at__isnull=True).order_by(
+        woking_user_stage_list = UserStage.objects.filter(user_line=self, stutas=0).order_by(
             'finish_at')
-        if woking_user_stage_list.count() == 1:
+        if woking_user_stage_list.count() >= 1:
             return woking_user_stage_list.last()
         else:
             return None
 
+    # 添加user_line下的所有stage
+    # user_line：要添加的user_line
+    # check：检查是否有未完成的stage
+    # 返回新添加的user_stage
+    def check_add_user_stage(self, check=True):
+        # 已经有未完成的stage
+        if check:
+            woking_stage = self.get_woking_user_stage()
+            if woking_stage:
+                return
+        return self.add_user_stage()
+
+    #     return
+
     # 添加下一个或第一个user_stage
     # 参数：check检查是否已经有同类型的已开启的user_stage。为true时，不添加，false时，强行再添加一个
-    def add_user_stage(self, check=True):
-        # 已经有未完成的stage
-        if check == True:
-            user_stage_list = UserStage.objects.filter(user_line__user=self.user, finish_at__isnull=True,
-                                                       user_line__line__line_type=self.line.line_type)
-            if user_stage_list.count() > 0:
-                return
+    # def check_add_user_stage(self, check=True):
+    #     # self.update_user_stage_status()
+    #     # 已经有未完成的stage
+    #     if check == True:
+    #         woking_stage = self.get_woking_user_stage()
+    #         if woking_stage:
+    #             return
+    #     self.add_user_stage()
 
-
-        # 找到第一个任务块
-        last_finished_user_stage = self.get_last_finished_user_stage()
-        if last_finished_user_stage:
-            last_stage = last_finished_user_stage.stage
+    # 添加user_stage 返回新添加的user_stagae
+    def add_user_stage(self):
+        # 找到第一个,或最后一个user_stage，包括已完成和未完成
+        if self.line.is_unlimited_line():
+            last_user_stage = self.get_last_user_stage(status_list=[1, 2])
+            if last_user_stage:
+                last_stage = last_user_stage.stage
+            else:
+                last_stage = None
+            next_stage = self.line.get_next_stage(last_stage)
+        elif self.line.is_week_line():
+            next_stage = Stage.get_week_stage()
         else:
-            last_stage = None
-        next_stage = self.line.get_next_stage(last_stage)
+            next_stage = None
+
+        # 如果该stage下还未添加mission，则不添加
+        # mission_list = next_stage.mission_set.all()
+        # if mission_list.count() == 0:
+        #     return
 
         # 添加阶段任务
         if next_stage:
@@ -285,13 +261,18 @@ class UserLine(models.Model):
             new_user_stage.user_line = self
             new_user_stage.stage = next_stage
             # 如果是周任务，添加结束时间
-            if self.line.line_type == 2:
+            if self.line.is_week_line():
                 start_dt, end_dt = get_current_week_range_datetime()
                 new_user_stage.start_at = start_dt
                 new_user_stage.end_at = end_dt
                 new_user_stage.name = start_dt.strftime("%Y") + '年第' + start_dt.strftime("%W") + "周任务"
             new_user_stage.save()
-        return
+        else:
+            return None
+        return new_user_stage
+
+
+USERSTAGE_STUTAS_CHOICES = ((0, '进行中'), (1, '已完成'), (2, '未完成'), (3, '可完成但未完成'))
 
 
 class UserStage(models.Model):
@@ -302,6 +283,7 @@ class UserStage(models.Model):
     start_at = models.DateTimeField(default=timezone.now, verbose_name='开始时间')
     finish_at = models.DateTimeField(null=True, verbose_name='完成时间')
     end_at = models.DateTimeField(null=True, verbose_name='结束时间')
+    stutas = models.IntegerField(default=0, verbose_name='状态')
 
     class Meta:
         db_table = 'bee_django_mission_user_stage'
@@ -317,9 +299,12 @@ class UserStage(models.Model):
             return self.name
         return self.stage.name
 
-    # 添加当前user_stage下的所有mission
-    def add_user_missions(self):
-        # 已经有未完成的stage
+    def check_add_user_mission(self, check=True):
+        self.add_user_mission()
+
+    # 添加当前user_stage下的所有mission，如果已添加，则不重复添加
+    def add_user_mission(self):
+        # 所有user_stage下的mission
         mission_list = Mission.objects.filter(stage=self.stage)
 
         for mission in mission_list:
@@ -331,50 +316,37 @@ class UserStage(models.Model):
                 new_user_misssion.user_stage = self
                 new_user_misssion.mission = mission
                 new_user_misssion.save()
+        return
+
+    # 更新该user_stage下，所有的mission完成情况
+    def update_user_mission(self):
+        # 所有未完成的user_mission
+        user_mission_list = self.usermission_set.filter(finish_at__isnull=True)
+        for user_mission in user_mission_list:
+            user_mission.get_user_mission_progress()
+        return
+
+    # 更新user_stage状态，检查进行中的，是否可以改变状态
+    def update_user_stage_status(self):
+        if self.stutas in [1, 2, 3]:
+            return self.stutas
 
-    # 检查stage是否完成
-    def check_finish(self):
         unfinish_list = UserMission.objects.filter(user_stage=self, finish_at__isnull=True)
-        if unfinish_list.count() > 0:
-            return False
-        return True
+        # 所有mission都已完成
+        if unfinish_list.count() == 0:
+            if self.user_line.line.auto_finish:
+                self.finish_at = get_now()
+                self.stutas = 1
+            else:
 
-    # 更新user_stage的完成情况，并根据情况，开启下一stage
-    # manual是否为手动完成
-    def update_user_stage_finish(self, manual=False):
-        # 如果任务线为不自动完成，则不更新
-        if not manual:
-            if not self.user_line.line.auto_finish:
-                return
-
-        res = self.check_finish()
-        if res:
-            self.finish_at = get_now()
-            self.save()
-            self.user_line.add_user_stage()
-
-            # 结束任务stage，并开启下一个stage
-            # def finish_and_start_next_user_stage(self):
-            #     try:
-            #         self.finish_at = datetime.datetime.now()
-            #         self.save()
-            #     except Exception as e:
-            #         print(e)
-            #
-            #     current_stage = self.stage
-            #     current_line = self.stage.line
-            #
-            #     # 有未完成stage，则退出
-            #     user_stage_list = UserStage.objects.filter(user=self.user, finish_at__isnull=True)
-            #     if user_stage_list.count() > 0:
-            #         return
-            #
-            #     # 找到下一个stage
-            #     next_stage_list = Stage.objects.filter(level__gt=current_stage.level, line=current_line)
-            #     if next_stage_list.count() > 0:
-            #         next_stage = next_stage_list.first()
-            #         # 添加到该用户
-            #         next_stage.add_to_user_stage(self.user)
+                self.stutas = 3
+        # 还有mission未完成
+        else:
+            # 如果是周任务，且过期
+            if self.user_line.line.is_week_line() and self.end_at < get_now():
+                self.stutas = 2
+        self.save()
+        return self.stutas
 
 
 class UserMission(models.Model):
@@ -418,7 +390,7 @@ class UserMission(models.Model):
         model_name = content_type.model
         app = apps.get_app_config(app_name)
         model = app.get_model(model_name)
-        print(app, model)
+        # print(app, model)
         # 查询条件
         user_field = content_type.user_field
 
@@ -452,7 +424,7 @@ class UserMission(models.Model):
                     value = condition.split(':')[1]
                     kwargs[key] = value
             # 周任务
-            if self.user_stage.user_line.line.line_type == 2:
+            if self.user_stage.user_line.line.is_week_line():
                 kwargs[timestamp_field + "__range"] = [self.user_stage.start_at, self.user_stage.end_at]
 
             # print(kwargs)
@@ -474,9 +446,6 @@ class UserMission(models.Model):
 
             else:
                 return 0
-            print(self.mission.name)
-            print(queryset)
-            print(count)
             # 比较
             if comparison_type == 1:
                 res = queryset["_agg"] >= count
@@ -484,35 +453,19 @@ class UserMission(models.Model):
                 res = queryset["_agg"] > count
             else:
                 res = None
-
+            # print(res)
             if res:
-                print('finished')
+                # print('finished')
                 self.finish_at = get_now()
                 self.save()
                 return 1
             else:
 
-                print(queryset["_agg"])
-                print('not finished')
+                # print(queryset["_agg"])
+                # print('not finished')
                 return round(queryset["_agg"].__int__(), 2) / count.__int__()
 
 
         except Exception as e:
             print(e)
         return 0
-
-
-# 创建user_line后，自动创建该用户对应的stage
-@receiver(post_save, sender=UserLine)
-def create_user_stage(sender, **kwargs):
-    user_line = kwargs['instance']
-    if kwargs['created']:
-        user_line.add_user_stage(check=True)
-
-
-# 创建user_stage后，自动创建该用户对应stage里的所有mission
-@receiver(post_save, sender=UserStage)
-def create_user_missions(sender, **kwargs):
-    user_stage = kwargs['instance']
-    if kwargs['created']:
-        user_stage.add_user_missions()
